from m4b_util.helpers import ffprobe


def check_output_folder(output_path, expected_files=None, check_func=None):
    """Check that expected output exists.

    :param output_path: Path to the directory to check.
    :param expected_files: Files we expect to find in output_path. If unspecified, defaults to ["segment_0000.mp3"].
    :param check_func: Function for checking each file in expected_files. Defaults to asserting ffprobe can read them.

    """
    # Default check function, in case one isn't passed in.
    def default_check(input_file_path):
        """Ensure ffprobe can read the file."""
        # Add in filename to make any failure more legible
        assert input_file_path.name and ffprobe.run_probe(input_file_path)

    # Set defaults
    if expected_files is None:
        expected_files = ["segment_0000.mp3"]
    if not check_func:
        check_func = default_check

    # Check the output
    for file_name in expected_files:
        file_path = output_path / file_name
        check_func(file_path)

    # Make sure there aren't any extra files in the directory
    for file in output_path.glob("*"):
        assert (file.name in expected_files)


def assert_file_path_is_file(input_file_path):
    """Ensure each file exists and is a file."""
    assert input_file_path.is_file()



"""Bind Command"""
import argparse
from pathlib import Path
import sys

from natsort import natsorted
from rich import print

from .Binder import Binder


def _parse_bind_args():
    """Parse all arguments."""
    parser = argparse.ArgumentParser(
        description="Take a folder of audio files and output an m4b.",
        prog="m4b-util bind"
    )
    parser.add_argument('input_folder', type=str, help="The folder of input files.")
    parser.add_argument('-a', "--author", type=str, help="Name of the author.")
    parser.add_argument('-c', "--cover", type=str, help="Image file to use as cover")
    parser.add_argument('-e', "--input-extension", type=str, action='extend', default=["mp3"], nargs="+",
                        help="Add a file extension to the allowed list used when grabbing input files, in addition "
                             "to mp3.")
    parser.add_argument('-o', "--output-dir", type=str, help="Directory to put the finished audiobook.")
    parser.add_argument('-n', "--output-name", type=str, help="Filename to use for finished audiobook. Default"
                                                              " is '[Author] - [Title].m4b'.")
    parser.add_argument('-t', "--title", type=str, help="Title of the audiobook.")
    parser.add_argument("--date", type=str, help="Date to include in metadata.")
    parser.add_argument("--decode-durations", "--decode-duration", action='store_true',
                        help="Fully decode each file to determine its duration (Slower, but more accurate).")
    parser.add_argument("--show-order", action='store_true',
                        help="Show the order the files would be read in, then exit.")
    parser.add_argument("--keep-temp-files", action='store_true', help="Skip cleanup. (Debugging)")
    parser.add_argument("--use-filename", "--use-filenames", action='store_true',
                        help="Use the filename as the chapter title instead of the title from the file's metadata.")

    return parser.parse_args(sys.argv[2:])


def run():
    """Entrypoint for bind subcommand."""
    binder = Binder()
    args = _parse_bind_args()

    # Make sure the output directory exists, if it was specified.
    if args.output_dir and not Path(args.output_dir).is_dir():
        print(f"[bold red]Error:[/] '{args.output_dir}' is not a directory.")
        return -1

    # Set info from args
    binder.author = args.author
    binder.cover = args.cover
    binder.output_dir = args.output_dir
    binder.output_name = args.output_name
    binder.title = args.title
    binder.date = args.date
    binder.decode_durations = args.decode_durations
    binder.keep_temp_files = args.keep_temp_files
    binder.use_filename = args.use_filename

    # Discover the input files
    audio_list = list()
    for ext in args.input_extension:
        # Remove leading dots, if any
        ext = ext.lstrip(".")

        # Glob them all!
        audio_list.extend(Path(args.input_folder).glob(f"*.{ext}"))
    input_files = natsorted(audio_list)

    # Print order, if applicable
    if args.show_order:
        i = 0
        for file in input_files:
            print(f"{i:3}:\t{file.name}")
            i += 1
        return 0

    # Add the files to the binder
    binder.files = input_files

    # Run the binder
    if binder.bind():
        # Tell the user where it is.
        print(f"[cyan]Writing '[yellow]{binder.output_path}[/]'")

    return 0



"""Test the split subcommand."""
from contextlib import contextmanager
import os
from pathlib import Path
from unittest import mock

import pytest

from m4b_util import split
from m4b_util.helpers import ffprobe
import testhelpers


@contextmanager
def change_cwd(new_path):
    """Change to a new directory, then change back when finished."""
    owd = Path.cwd()  # Original Working Directory
    os.chdir(new_path)
    try:
        yield
    finally:
        os.chdir(owd)


def _run_split_cmd(arg_list):
    """Patch the runtime arguments, then run the split command."""
    argv_patch = ["m4b-util", "split"]
    argv_patch.extend(arg_list)

    with mock.patch("sys.argv", argv_patch):
        split.run()


@mock.patch("m4b_util.split.subcommand.SilenceFinder")
def test_silence_modes(finder_mock):
    """Run in silence-split mode."""
    finder_mock = finder_mock()
    finder_mock.find.return_value = None
    for mode in ["s", "silence", "silences"]:
        with pytest.raises(SystemExit) as e:
            _run_split_cmd([mode, "Not-really-a-file"])
        assert e.value.code == 1
        finder_mock.find.assert_called_once()
        finder_mock.reset_mock()


@mock.patch("m4b_util.split.subcommand.ChapterFinder")
def test_chapter_modes(finder_mock):
    """Run in chapter-split mode."""
    finder_mock = finder_mock()
    finder_mock.find.return_value = None
    for mode in ["c", "chapter", "chapters"]:
        with pytest.raises(SystemExit) as e:
            _run_split_cmd([mode, "Not-really-a-file"])
        assert e.value.code == 1
        finder_mock.find.assert_called_once()
        finder_mock.reset_mock()


def test_unexpected_mode(capsys):
    """Alert the user if an unexpected mode is requested."""
    with pytest.raises(SystemExit) as e:
        _run_split_cmd(["not-a-real-mode", "Not-really-a-file"])
    assert e.value.code == 1
    output = capsys.readouterr()
    assert "Unexpected mode" in output.out



def test_split_silence(tmp_path, silences_file_path):
    """Split a file on silence, default settings."""
    output_path = tmp_path / "output"
    output_path.mkdir()
    expected_files = [
        "segment_0000.mp3",
        "segment_0001.mp3",
        "segment_0002.mp3",
        "segment_0003.mp3",
    ]
    with change_cwd(output_path):
        _run_split_cmd(["silence", str(silences_file_path), "--silence-duration", "1.0" ])
        testhelpers.check_output_folder(output_path, expected_files)


def test_split_chapter(tmp_path, chaptered_audio_file_path):
    """Split a file on silence, default settings."""
    output_path = tmp_path / "output"
    output_path.mkdir()
    expected_files = [
        "segment_0000.mp3",
        "segment_0001.mp3",
        "segment_0002.mp3",
        "segment_0003.mp3",
        "segment_0004.mp3",
        "segment_0005.mp3",
        "segment_0006.mp3",
        "segment_0007.mp3",
    ]
    with change_cwd(output_path):
        _run_split_cmd(["chapter", str(chaptered_audio_file_path)])
        testhelpers.check_output_folder(output_path, expected_files)


def test_split_audio_start_end_times(tmp_path, chaptered_audio_file_path):
    """Split a file, with start and end times specified."""
    output_path = tmp_path / "output"
    _run_split_cmd([
            "chapter",
            str(chaptered_audio_file_path),
            "-o", str(output_path),
            "-s", "1.75",
            "-e", "5.03"
        ])
    testhelpers.check_output_folder(output_path)


def test_split_audio_fail(tmp_path, capsys):
    """Exit with non-zero code when asked to process a non-audio file."""
    fake_file_path = tmp_path / "not-a-real-file.m4a"
    open(fake_file_path, 'a').close()
    with pytest.raises(SystemExit) as e:
        _run_split_cmd(["silence", str(fake_file_path), "-o", str(tmp_path)])
    assert (e.value.code == 1)
    output = capsys.readouterr()
    assert "Could not determine segment times." in output.out


def test_split_audio_duration(tmp_path, silences_file_path):
    """Adjust silence duration limits."""
    output_path = tmp_path / "output"
    cmd = [
        "silence",
        str(silences_file_path),
        "-o", str(output_path),
        "--silence-duration", "1.5"
    ]
    expected_files = (
        "segment_0000.mp3",
        "segment_0001.mp3",
        "segment_0002.mp3",
        "segment_0003.mp3",
    )
    _run_split_cmd(cmd)
    testhelpers.check_output_folder(output_path, expected_files)


def test_split_audio_threshold_lower(tmp_path, variable_volume_segments_file_path, capsys):
    """Adjust silence threshold limits to a threshold lower than any noise in the file."""
    output_path = tmp_path / "output"
    cmd = [
        "silence",
        str(variable_volume_segments_file_path),
        "-o", str(output_path),
        "--silence-threshold", "-45",
        "--silence-duration", "1.5"
    ]
    with pytest.raises(SystemExit) as e:
        _run_split_cmd(cmd)
    assert (e.value.code == 1)
    output = capsys.readouterr()
    assert "No segments found." in output.out


def test_split_audio_threshold_higher(tmp_path, variable_volume_segments_file_path, capsys):
    """Adjust silence threshold limits to a threshold higher than any noise in the file."""
    output_path = tmp_path / "output"
    cmd = [
        "silence",
        str(variable_volume_segments_file_path),
        "-o", str(output_path),
        "--silence-threshold", "-5",
        "--silence-duration", "1.5"
    ]
    with pytest.raises(SystemExit) as e:
        _run_split_cmd(cmd)
    assert (e.value.code == 1)
    output = capsys.readouterr()
    assert "Not enough segments found." in output.out


def test_split_audio_threshold(tmp_path, variable_volume_segments_file_path):
    """Adjust silence threshold limits to a threshold right between the two volume levels in the file."""
    output_path = tmp_path / "output"
    cmd = [
        "silence",
        str(variable_volume_segments_file_path),
        "-o", str(output_path),
        "--silence-threshold", "-20",
        "--silence-duration", "1.5"
    ]
    expected_files = (
        "segment_0000.mp3",
        "segment_0001.mp3",
        "segment_0002.mp3",
        "segment_0003.mp3",
    )
    _run_split_cmd(cmd)
    testhelpers.check_output_folder(output_path, expected_files)


def test_split_audio_segment_pattern(tmp_path, silences_file_path):
    """Specify a non-default segment pattern."""
    output_path = tmp_path / "output"
    cmd = [
        "silence",
        str(silences_file_path),
        "-o", str(output_path),
        "--silence-duration", "1.5",
        "-p", "chunk_{:02d}.mp3",
    ]
    # Make sure the files we expect are there
    expected_files = (
        "chunk_00.mp3",
        "chunk_01.mp3",
        "chunk_02.mp3",
        "chunk_03.mp3",
    )
    _run_split_cmd(cmd)
    testhelpers.check_output_folder(output_path, expected_files)



import pytest

from m4b_util.helpers import cover_utils


def test_extract_cover(tmp_path, covered_audio_file):
    """Grab a cover."""
    output_extensions = ["png", "jpg", "jpeg", "JPG"]
    for extension in output_extensions:
        output = tmp_path / f"cover.{extension}"
        cover_utils.extract_cover(covered_audio_file, output)
        assert output.is_file()


def test_extract_cover_invalid_output_extension(tmp_path, covered_audio_file):
    """Reject invalid output extensions."""
    output = tmp_path / "cover.bad_ext"
    with pytest.raises(ValueError) as e:
        cover_utils.extract_cover(covered_audio_file, output)
    assert "Output extension must be one of" in str(e)


def test_extract_cover_non_covered_file(tmp_path, mp3_file_path, capsys):
    """Handle files without a cover image."""
    output = tmp_path / "cover.png"
    cover_utils.extract_cover(mp3_file_path, output)
    sysout = capsys.readouterr()
    assert "Unable to extract cover" in sysout.out
    assert "ffmpeg" not in sysout.out
    assert not output.exists()



"""Fixtures for expected data."""
import pytest


@pytest.fixture(scope='session')
def expected_data():
    """Dictionary of data expected after processing the tone files.

    For some reason, the mp3 timestamps are different from wav and m4a.
    """
    return {
        "default": {
            "metadata": (';FFMETADATA1\nmajor_brand=M4A\nminor_version=512\ncompatible_brands=M4A isomiso2\n'
                         'title=None\n'
                         'artist=None\n'
                         'album=None\n'
                         'date=None\n'
                         'genre=Audiobook\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=0\nEND=4999\ntitle=1\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=5000\nEND=9999\ntitle=2\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=10000\nEND=14999\ntitle=3\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=15000\nEND=19999\ntitle=4\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=20000\nEND=24999\ntitle=5\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=25000\nEND=29999\ntitle=6\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=30000\nEND=34999\ntitle=7\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=35000\nEND=39999\ntitle=8\n'
                         ),
            "chapters": [
                {'startTime': 0, 'title': 1},
                {'startTime': 5000, 'title': 2},
                {'startTime': 10000, 'title': 3},
                {'startTime': 15000, 'title': 4},
                {'startTime': 20000, 'title': 5},
                {'startTime': 25000, 'title': 6},
                {'startTime': 30000, 'title': 7},
                {'startTime': 35000, 'title': 8},
            ]
        },
        "mp3": {
            "metadata": (';FFMETADATA1\nmajor_brand=M4A\nminor_version=512\ncompatible_brands=M4A isomiso2\n'
                         'title=None\n'
                         'artist=None\n'
                         'album=None\n'
                         'date=None\n'
                         'genre=Audiobook\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=0\nEND=5039\ntitle=1\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=5040\nEND=10079\ntitle=2\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=10080\nEND=15119\ntitle=3\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=15120\nEND=20159\ntitle=4\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=20160\nEND=25199\ntitle=5\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=25200\nEND=30239\ntitle=6\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=30240\nEND=35279\ntitle=7\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=35280\nEND=40319\ntitle=8\n'
                         ),
            "chapters": [
                {'startTime': 0, 'title': 1},
                {'startTime': 5040, 'title': 2},
                {'startTime': 10080, 'title': 3},
                {'startTime': 15120, 'title': 4},
                {'startTime': 20160, 'title': 5},
                {'startTime': 25200, 'title': 6},
                {'startTime': 30240, 'title': 7},
                {'startTime': 35280, 'title': 8},
            ]
        },
        "use_filename": {
            "metadata": (';FFMETADATA1\nmajor_brand=M4A\nminor_version=512\ncompatible_brands=M4A isomiso2\n'
                         'title=None\n'
                         'artist=None\n'
                         'album=None\n'
                         'date=None\n'
                         'genre=Audiobook\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=0\nEND=5039\ntitle=1 - 110Hz\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=5040\nEND=10079\ntitle=2 - 220Hz\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=10080\nEND=15119\ntitle=3 - 330Hz\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=15120\nEND=20159\ntitle=4 - 440Hz\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=20160\nEND=25199\ntitle=5 - 550Hz\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=25200\nEND=30239\ntitle=6 - 660Hz\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=30240\nEND=35279\ntitle=7 - 770Hz\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=35280\nEND=40319\ntitle=8 - 880Hz\n'
                         ),
            "chapters": [
                {'startTime': 0, 'title': "1 - 110Hz"},
                {'startTime': 5040, 'title': "2 - 220Hz"},
                {'startTime': 10080, 'title': "3 - 330Hz"},
                {'startTime': 15120, 'title': "4 - 440Hz"},
                {'startTime': 20160, 'title': "5 - 550Hz"},
                {'startTime': 25200, 'title': "6 - 660Hz"},
                {'startTime': 30240, 'title': "7 - 770Hz"},
                {'startTime': 35280, 'title': "8 - 880Hz"},
            ]
        }
    }



"""The Split pacakge."""
from .subcommand import run  # noqa: F401
from .SilenceFinder import SilenceFinder  # noqa: F401
from .Splitter import Splitter
from .ChapterFinder import ChapterFinder  # noqa: F401
from .SegmentData import SegmentData  # noqa: F401



"""PyTest Configuration."""
from pathlib import Path
from subprocess import run

import pytest

from expected_data import expected_data  # noqa

# Let pytest know we would like them to handle asserts in our helper code
pytest.register_assert_rewrite("testhelpers")


def _generate_tone_files(dir_path, extension):
    """Generate a folder of sine wave audio files.

    Creates 8 files with 2.5 seconds of pure sine wave at the base frequency, increasing by the base frequency
    each time.

    :param dir_path: Base folder in which to place the output.
    :param extension: File extension (and thus, implicitly, the format) to use for the output files.
    """
    base_frequency = 110  # Hz
    for i in range(1, 9):
        frequency = base_frequency * i
        output_path = dir_path / f"{i} - {frequency}Hz.{extension}"
        cmd = ["ffmpeg", "-f", "lavfi", "-i", f"sine=frequency={frequency}:sample_rate=48000:duration=2.5",
               "-filter_complex", "[0]apad=pad_dur=2.5[s0]", "-map", "[s0]", output_path]
        run(cmd, capture_output=True, check=True)


@pytest.fixture(scope='session')
def test_data_path():
    """Path to the test data directory."""
    # Since __file__ returns the file name, we need to call parent to get the directory
    return Path(__file__).parent.joinpath("data").absolute()


@pytest.fixture()
def wav_path(tmp_path):
    """Path to a temporary folder with a number of short wav files."""
    wav_path = tmp_path / "wav"
    wav_path.mkdir()
    _generate_tone_files(wav_path, "wav")

    return wav_path


@pytest.fixture()
def mp3_path(tmp_path):
    """Path to a temporary folder with a number of short mp3 files."""
    mp3_path = tmp_path / "mp3"
    mp3_path.mkdir()
    _generate_tone_files(mp3_path, "mp3")

    return mp3_path


@pytest.fixture()
def m4a_path(tmp_path):
    """Path to a temporary folder with a number of short m4a files."""
    m4a_path = tmp_path / "m4a"
    m4a_path.mkdir()
    _generate_tone_files(m4a_path, "m4a")

    return m4a_path


@pytest.fixture()
def covered_audio_file(tmp_path, test_data_path):
    """Path to an m4a file with a cover image."""
    output_path = tmp_path / "covered_audio_file.m4a"
    cover_path = test_data_path / "cover.png"
    cmd = ["ffmpeg", "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5",
           "-i", cover_path, "-c:v", "png",
           "-map", "0:a", "-map", "1", "-disposition:v:0", "attached_pic",
           output_path
           ]
    run(cmd, capture_output=True, check=True)
    return output_path


@pytest.fixture()
def video_only_file(tmp_path):
    """File containing a single video stream test pattern."""
    output_path = tmp_path / "video_only.mp4"
    cmd = ["ffmpeg", "-f", "lavfi", "-i", "testsrc=duration=5.3:size=qcif:rate=10", output_path]
    run(cmd, capture_output=True, check=True)
    return output_path


@pytest.fixture()
def mp3_file_path(tmp_path):
    """A single mp3 file. 2.5s tone, 2.5s silence."""
    output_path = tmp_path / "mp3.mp3"
    cmd = ["ffmpeg", "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5",
           "-filter_complex", "[0]apad=pad_dur=2.5[s0]", "-map", "[s0]", output_path]
    run(cmd, capture_output=True, check=True)
    return output_path



"""Test the Binder class."""
from pathlib import Path
import shutil
from subprocess import run
from tempfile import TemporaryDirectory

from natsort import natsorted
import pytest

from m4b_util.bind.Binder import Binder
from m4b_util.helpers import ffprobe
import testhelpers


def _do_scan(binder, files, **kwargs):
    """Use binder to scan files, then test the binder's attributes based on kwargs."""
    binder.files = natsorted(files)
    binder._scan_files(binder.files)  # noqa
    for attr_name, attr_value in kwargs.items():
        assert (getattr(binder, attr_name) == attr_value)


def _add_metadata(file_path, **kwargs):
    """Add metadata specified by kwargs to file_path."""
    with TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir) / f"temp{file_path.suffix}"
        cmd = ["ffmpeg", "-i", file_path]

        # Add in metadata from kwargs
        for key, value in kwargs.items():
            cmd.append("-metadata")
            cmd.append(f"{key}={value}")

        # Add output path
        cmd.append(tmp_path)

        run(cmd, capture_output=True, check=True)

        shutil.move(tmp_path, file_path)


def test_output_path():
    """Set the output path in various ways."""
    b = Binder()
    assert (str(b.output_path) == "None - None.m4b")  # Default
    # Author
    b.author = "Write-y McAuthor"
    assert (str(b.output_path) == "Write-y McAuthor - None.m4b")
    # Author + Title
    b.title = "Test Book: An exercise of m4b-util"
    assert (str(b.output_path) == "Write-y McAuthor - Test Book: An exercise of m4b-util.m4b")
    # Dir + Author + Title
    b.output_dir = "Fake Dir"
    assert (str(b.output_path) == "Fake Dir/Write-y McAuthor - Test Book: An exercise of m4b-util.m4b")
    # Dir + Name (Overwrites Author & Title), Adds m4b extension
    b.output_name = "Author McWrite-y"
    assert (str(b.output_path) == "Fake Dir/Author McWrite-y.m4b")
    # Dir + Name (Overwrites Author & Title), does not double up m4b extension
    b.output_name = "Author McWrite-y.m4b"
    assert (str(b.output_path) == "Fake Dir/Author McWrite-y.m4b")


def test_file_scanner_mp3(mp3_path, expected_data):
    """Get the correct data from a set of mp3 files."""
    _do_scan(
        Binder(),
        mp3_path.glob("*"),
        title=None,
        author=None,
        date=None,
        chapters=expected_data['mp3']['chapters'],
        metadata=expected_data['mp3']['metadata']
    )


def test_file_scanner_wav(wav_path, expected_data):
    """Get the correct data from a set of mp3 files."""
    _do_scan(
        Binder(),
        wav_path.glob("*"),
        title=None,
        author=None,
        date=None,
        chapters=expected_data['default']['chapters'],
        metadata=expected_data['default']['metadata']
    )


def test_file_scanner_m4a(m4a_path, expected_data):
    """Get the correct data from a set of mp3 files."""
    _do_scan(
        Binder(),
        m4a_path.glob("*"),
        title=None,
        author=None,
        date=None,
        chapters=expected_data['default']['chapters'],
        metadata=expected_data['default']['metadata']
    )


def test_file_scanner_use_filenames(mp3_path, expected_data):
    """Use filenames instead of metadata when scanning files."""
    b = Binder()
    b.use_filename = True
    _do_scan(
        b,
        mp3_path.glob("*"),
        title=None,
        author=None,
        date=None,
        chapters=expected_data['use_filename']['chapters'],
        metadata=expected_data['use_filename']['metadata']
    )


def test_metadata_from_input(mp3_path):
    """Pick up metadata from our input files."""
    file = mp3_path / "2 - 220Hz.mp3"
    _add_metadata(file, album="Book Title", artist="Author McWrite-y", date="2022")
    _do_scan(
        Binder(),
        mp3_path.glob("*"),
        title="Book Title",
        author="Author McWrite-y",
        date="2022"
    )


def test_grab_first_metadata_only(mp3_path):
    """Only fill metadata items the first time we see them."""
    file = mp3_path / "2 - 220Hz.mp3"
    _add_metadata(file, album="Book Title", artist="Author McWrite-y")
    second_file = mp3_path / "3 - 330Hz.mp3"
    _add_metadata(second_file, album="Fake Book Title", artist="Not a Real Author")
    _do_scan(
        Binder(),
        mp3_path.glob("*"),
        title="Book Title",
        author="Author McWrite-y"
    )


def test_prefer_user_supplied_metadata(mp3_path):
    """Don't fill metadata from files if we already have it."""
    file = mp3_path / "2 - 220Hz.mp3"
    _add_metadata(file, album="Fake Book Title", artist="Not Author McWrite-y", date="1011")
    b = Binder()
    b.title = "Book Title"
    b.author = "Author McWrite-y"
    b.date = "2022"
    _do_scan(
        b,
        mp3_path.glob("*"),
        title="Book Title",
        author="Author McWrite-y",
        date="2022"
    )


def test_scanning_nonaudio_file(mp3_path, capsys):
    """Print a warning if we can't parse a file."""
    # Create a blank mp3 file
    open(mp3_path / "not-really-audio.mp3", 'a').close()
    _do_scan(Binder(), mp3_path.glob("*"))
    output = capsys.readouterr()
    assert ("Warning: Unable to parse" in output.out)


def test_status_updates_during_scan(wav_path, capsys):
    """Show the user what we are doing while we scan."""
    _do_scan(Binder(), wav_path.glob("*"))
    output = capsys.readouterr()
    assert (output.out == "Collecting file data...\nFile scan complete.\n")


def test_concatenation(m4a_path):
    """Concatenate all the files into one."""
    b = Binder()
    b.temp_files = natsorted(m4a_path.glob("*"))
    out_file_path = b._concatenate_all_audio_files()

    probe = ffprobe.run_probe(out_file_path)
    assert probe
    assert probe.format['duration'] == "40.022000"
    assert probe.audio['duration'] == "40.021333"
    assert not probe.tags.get('title')


def test_filenames_with_single_quotes(m4a_path):
    """Create a well-formed file when one of the inputs has quotes in its name."""
    orig_file = m4a_path / "2 - 220Hz.m4a"
    quote_file = m4a_path / "2 - '220Hz'.m4a"
    shutil.move(orig_file, quote_file)

    b = Binder()
    b.temp_files = natsorted(m4a_path.glob("*"))
    out_file_path = b._concatenate_all_audio_files()

    probe = ffprobe.run_probe(out_file_path)
    assert probe
    assert probe.format['duration'] == "40.022000"
    assert probe.audio['duration'] == "40.021333"  # If the file is broken, this will be shorter
    assert not probe.tags.get('title')


def test_decode_durations(wav_path):
    """Read the duration of a file by decoding it."""
    file = wav_path / "4 - 440Hz.wav"
    b = Binder()
    meta_duration = b._get_duration(file)
    b.decode_durations = True
    decoded_duration = b._get_duration(file)
    assert (meta_duration == decoded_duration == 5000)


def test_decode_duration_fake_file(tmp_path):
    """Read the duration of a file by decoding it."""
    file = tmp_path / "not-really-audio.mp3"
    open(file, 'a').close()
    b = Binder()
    b.decode_durations = True
    with pytest.raises(RuntimeError) as e:
        b._get_duration(file)
    assert ("Could not get audio stream." in str(e.value))


def test_get_duration_no_metadata(tmp_path, capsys):
    """Throw an error if no duration metadata, such as in mkv's."""
    # Create the file
    file = tmp_path / "no_duration.mkv"
    cmd = ["ffmpeg", "-f", "lavfi", "-i", "sine=frequency=110:sample_rate=48000:duration=2.5",
           "-filter_complex", "[0]apad=pad_dur=2.5[s0]", "-map", "[s0]", file]
    run(cmd, capture_output=True, check=True)

    # Make sure it reads successfully if decode_durations is set
    b = Binder()
    b.decode_durations = True
    assert (b._get_duration(file) == 5000)

    # Make sure it doesn't work when decode_durations is not set
    b.decode_durations = False
    with pytest.raises(RuntimeError) as e:
        b._get_duration(file)
    assert ("Cannot parse duration listed in file." in str(e.value))

    # Make sure it gets handled and mentioned to user during scanning
    _do_scan(Binder(), [file])
    output = capsys.readouterr()
    assert ("Failed to determine duration" in output.out)


def test_add_chapter_info(m4a_path):
    """Add chapter info to a file."""
    file = m4a_path / "4 - 440Hz.m4a"
    b = Binder()
    b.chapters.append({"title": "Sound", "startTime": 0})
    b.chapters.append({"title": "Silence", "startTime": 2500})
    out_file_path = b._add_chapter_info(file)

    probe = ffprobe.run_probe(out_file_path)
    assert probe
    assert len(probe.chapters) == 2
    assert probe.chapters[0]['tags']['title'] == "Sound"
    assert probe.chapters[1]['tags']['title'] == "Silence"


def test_add_cover(m4a_path, test_data_path):
    """Add chapter info to a file."""
    file = m4a_path / "4 - 440Hz.m4a"
    b = Binder()
    b.cover = test_data_path / "cover.png"
    out_file_path = b._add_cover(file)

    probe = ffprobe.run_probe(out_file_path)
    assert probe
    assert probe.streams[1]['codec_name'] == "png"


def test_ffmpeg_fail():
    """Handle it when an ffmpeg run fails."""
    b = Binder()
    with pytest.raises(SystemExit) as e:
        b._run_ffmpeg(["ffmpeg", "-invalid_option"], "Make ffmpeg fail.")
    assert (e.value.code == 1)


def test_ffmpeg_fail_keep_temp_files():
    """Handle it when an ffmpeg run fails."""
    b = Binder()
    b.keep_temp_files = True
    canary = Path(b.temp_path) / "canary"
    open(canary, 'a').close()
    with pytest.raises(SystemExit) as e:
        b._run_ffmpeg(["ffmpeg", "-invalid_option"], "Make ffmpeg fail.")
    assert (e.value.code == 1)
    expected_files = ["canary"]
    testhelpers.check_output_folder(b.temp_path, expected_files, check_func=testhelpers.assert_file_path_is_file)


def test_bind_no_files(capsys):
    """Show an error when there are no files to bind."""
    b = Binder()
    assert (b.bind() is False)
    output = capsys.readouterr()
    assert ("No input files found" in output.out)


def test_bind_keep_temp_files(mp3_path, tmp_path, test_data_path, capsys):
    """Keep all temp files when asked."""
    out_file_path = tmp_path / "Book.m4b"
    b = Binder()
    b.files = natsorted(mp3_path.glob("*"))
    b.cover = test_data_path / "cover.png"
    b.keep_temp_fiyyles = True
    b.output_name = str(out_file_path)
    b.bind()

    # Make sure we told the user where the temp dir is
    output = capsys.readouterr()
    assert ("Temp files can be found at" in output.out)

    # Make sure the files we expect are there
    expected_files = (
        "1 - 110Hz.m4a",
        "2 - 220Hz.m4a",
        "3 - 330Hz.m4a",
        "4 - 440Hz.m4a",
        "5 - 550Hz.m4a",
        "6 - 660Hz.m4a",
        "7 - 770Hz.m4a",
        "8 - 880Hz.m4a",
        "long.m4a",
        "filelist",
        "coverless.m4b",
        "finished.m4b",
        "ffmetadata",
    )
    testhelpers.check_output_folder(b.temp_path, expected_files, check_func=testhelpers.assert_file_path_is_file)


def test_bind_coverless(mp3_path, tmp_path):
    """Create an audiobook without a cover."""
    out_file_path = tmp_path / "Book.m4b"
    b = Binder()
    b.files = natsorted(mp3_path.glob("*"))
    b.output_name = str(out_file_path)
    b.bind()

    # Verify output
    probe = ffprobe.run_probe(out_file_path)
    assert probe

    # Check chapters
    assert len(probe.chapters) == 8
    assert probe.chapters[0]['tags']['title'] == "1"
    assert probe.chapters[1]['tags']['title'] == "2"
    assert probe.chapters[2]['tags']['title'] == "3"
    assert probe.chapters[3]['tags']['title'] == "4"
    assert probe.chapters[4]['tags']['title'] == "5"
    assert probe.chapters[5]['tags']['title'] == "6"
    assert probe.chapters[6]['tags']['title'] == "7"
    assert probe.chapters[7]['tags']['title'] == "8"

    # Check duration
    assert probe.format['duration'] == "40.022000"
    assert probe.audio['duration'] == "40.021333"

    # Check metadata
    assert probe.tags['title'] == "None"
    assert probe.tags['artist'] == "None"
    assert probe.tags['album'] == "None"
    assert probe.tags['date'] == "None"
    assert probe.tags['genre'] == "Audiobook"


def test_bind_covered(mp3_path, tmp_path, test_data_path):
    """Create an audiobook with a cover."""
    b = Binder()
    b.files = natsorted(mp3_path.glob("*"))
    b.output_dir = str(tmp_path)
    b.output_name = "Book.m4b"
    b.cover = test_data_path / "cover.png"
    b.bind()

    # Verify output
    probe = ffprobe.run_probe(tmp_path / "Book.m4b")
    assert probe

    # Check chapters
    assert len(probe.chapters) == 8
    assert probe.chapters[0]['tags']['title'] == "1"
    assert probe.chapters[1]['tags']['title'] == "2"
    assert probe.chapters[2]['tags']['title'] == "3"
    assert probe.chapters[3]['tags']['title'] == "4"
    assert probe.chapters[4]['tags']['title'] == "5"
    assert probe.chapters[5]['tags']['title'] == "6"
    assert probe.chapters[6]['tags']['title'] == "7"
    assert probe.chapters[7]['tags']['title'] == "8"

    # Check duration
    assert probe.format['duration'] == "40.022000"
    assert probe.audio['duration'] == "40.021333"

    # Check metadata
    assert probe.tags['title'] == "None"
    assert probe.tags['artist'] == "None"
    assert probe.tags['album'] == "None"
    assert probe.tags['date'] == "None"

    # Check cover
    assert probe.streams[2]['codec_name'] == "png"



"""Test the Bind subcommand."""
from unittest.mock import patch

import m4b_util


def _run_bind_cmd(arg_list):
    argv_patch = ["m4b-util", "bind"]
    argv_patch.extend(arg_list)

    with patch("sys.argv", argv_patch):
        m4b_util.bind.run()


def test_show_order(mp3_path, capsys):
    """Show the order the files would be bound in, if asked."""
    _run_bind_cmd([str(mp3_path), "--show-order"])

    output = capsys.readouterr()
    expected_output = ('  0:    1 - 110Hz.mp3\n'
                       '  1:    2 - 220Hz.mp3\n'
                       '  2:    3 - 330Hz.mp3\n'
                       '  3:    4 - 440Hz.mp3\n'
                       '  4:    5 - 550Hz.mp3\n'
                       '  5:    6 - 660Hz.mp3\n'
                       '  6:    7 - 770Hz.mp3\n'
                       '  7:    8 - 880Hz.mp3\n'
                       )
    assert (output.out == expected_output)


def test_nonexistent_output_dir(capsys):
    """Alert the user if the output dir doesn't exist."""
    _run_bind_cmd(["fake/input/dir", "--output-dir", "/some/nonexistent/directory/"])

    output = capsys.readouterr()
    assert ("is not a directory" in output.out)


def test_bind_wav_defaults(wav_path, tmp_path, capsys):
    """Bind a folder of wav files into and audiobook."""
    _run_bind_cmd([str(wav_path), "-e", "wav", "-o", str(tmp_path)])

    # Check for output
    output_path = tmp_path / "None - None.m4b"
    assert output_path.is_file()

    probe = m4b_util.helpers.ffprobe.run_probe(output_path)
    assert probe
    assert probe.format['duration'] == "40.022000"
    assert probe.audio['duration'] == "40.021333"
    assert probe.tags['title'] == "None"
    assert probe.tags['artist'] == "None"
    assert probe.tags['album'] == "None"
    assert probe.tags['date'] == "None"
    assert probe.tags['genre'] == "Audiobook"

    # Make sure we tell the user where the output file is. Since rich auto-wraps the output unpredictably, we will
    # just check to make sure the start of the message, and the book file name are present.
    output = capsys.readouterr()
    assert ("Writing " in output.out)
    assert ("None - None.m4b" in output.out)


def test_bind_fail(tmp_path, capsys):
    """Avoid telling the user we output a file when we didn't."""
    fake_folder = tmp_path / "fake"
    fake_folder.mkdir()
    _run_bind_cmd([str(fake_folder)])
    output = capsys.readouterr()
    assert "Writing" not in output.out



from m4b_util.split import ChapterFinder, SegmentData


def test_find_chapters(chaptered_audio_file_path):
    """Read the chapter metadata into an object."""
    expected = [
        SegmentData(id=0, start_time=00.0, end_time=02.5, title="110Hz - Loud"),
        SegmentData(id=1, start_time=02.5, end_time=05.0, title="110Hz - Soft"),
        SegmentData(id=2, start_time=05.0, end_time=07.5, title="220Hz - Loud"),
        SegmentData(id=3, start_time=07.5, end_time=10.0, title="220Hz - Soft"),
        SegmentData(id=4, start_time=10.0, end_time=12.5, title="330Hz - Loud"),
        SegmentData(id=5, start_time=12.5, end_time=15.0, title="330Hz - Soft"),
        SegmentData(id=6, start_time=15.0, end_time=17.5, title="440Hz - Loud"),
        SegmentData(id=7, start_time=17.5, end_time=19.999, title="440Hz - Soft"),
    ]
    assert ChapterFinder(chaptered_audio_file_path).find() == expected


def test_start_end_times(chaptered_audio_file_path):
    """Filter out chapters based on a custom start and end time."""
    expected = [
        SegmentData(id=2, start_time=05.0, end_time=07.5, title="220Hz - Loud"),
        SegmentData(id=3, start_time=07.5, end_time=10.0, title="220Hz - Soft"),
        SegmentData(id=4, start_time=10.0, end_time=12.5, title="330Hz - Loud"),
    ]
    assert ChapterFinder(
        chaptered_audio_file_path,
        start_time=2.75,
        end_time=14.0
    ).find() == expected


def test_find_chapters_no_chapters(mp3_path):
    """Read an audio file that has no chapter data."""
    assert ChapterFinder(mp3_path / "1 - 110Hz.mp3").find() == []


def test_find_chapters_unreadable_file(tmp_path):
    """Find no chapters in a non-audio file."""
    fakefile = tmp_path / "Not-a-real-file.mp3"
    open(fakefile, 'a').close()
    assert ChapterFinder(fakefile).find() == []



"""ParallelFFmpeg Tests."""
from time import sleep
from unittest import mock

import pytest

from m4b_util.helpers.ParallelFFmpeg import ParallelFFmpeg


@pytest.fixture()
@mock.patch("m4b_util.helpers.ParallelFFmpeg.Progress")
def processor(progress):
    """Set up the processor and mock out the progress tracker."""
    p = ParallelFFmpeg("Testing")
    p._tasklist.append({  # noqa
        "name": "Master",
        "task_id": 0
    })
    p._tasklist.append({  # noqa
        "name": "Frist",
        "task_id": None
    })

    return p, progress()


def _add_to_q(q, message):
    """Add message to queue, and wait until the queue is not empty."""
    q.put(message)
    while q.empty():
        sleep(0.25)  # Give time for the queue to populate


def test_read_status_q_empty(processor):
    """Read an empty queue idempotently."""
    p, progress = processor
    p._read_status_queue()  # Should be a no-op, since queue is empty
    assert p._tasklist[1]["task_id"] is None
    progress.update.assert_not_called()


def test_read_status_q_new(processor):
    """Read a new task event."""
    # Set things up
    p, progress = processor
    progress.add_task = mock.MagicMock(return_value=333)
    _add_to_q(p._status_q, (1, "started"))

    # Test it!
    p._read_status_queue()
    assert p._tasklist[1]["task_id"] == 333
    progress.add_task.assert_called()
    progress.update.assert_not_called()


def test_read_status_q_update(processor):
    """Read a progress update."""
    p, progress = processor
    p._tasklist[1]["task_id"] = 24

    # 50% complete
    _add_to_q(p._status_q, (1, 50))
    p._read_status_queue()
    progress.update.assert_called_with(24, completed=50)

    # 100% Complete
    _add_to_q(p._status_q, (1, 100))
    p._read_status_queue()
    progress.update.assert_called_with(24, completed=100)


def test_read_status_finished(processor):
    """Read a task finished event."""
    p, progress = processor
    p._tasklist[1]["task_id"] = 48

    _add_to_q(p._status_q, (1, "finished"))
    p._read_status_queue()
    progress.update.assert_has_calls([
        mock.call(48, completed=100, visible=False),  # Specified Task
        mock.call(0, advance=1)  # Master Task
    ])


def test_read_status_failed(processor):
    """Read a task failed event."""
    p, progress = processor
    p._tasklist[1]["task_id"] = 12

    _add_to_q(p._status_q, (1, "failed"))
    p._read_status_queue()
    progress.console.print.assert_called_with("[red]Error:[/] Failed to process Frist")
    progress.update.assert_has_calls([
        mock.call(12, visible=False),  # Specified Task
        mock.call(0, advance=1)  # Master Task
    ])


@mock.patch("m4b_util.helpers.ParallelFFmpeg.FFProgress")
def test_ffmpeg_job(ff, processor):
    """Run a single worker job."""
    # Mock out the run command so we don't have to be dependent on system configurations.
    ff().run.return_value = [0, 50, 100]
    p, _ = processor
    status_q = mock.MagicMock()

    # Add jobs
    p._input_q.put((60, ["should-pass"]))
    p._input_q.put((72, ["should-pass"]))
    p._input_q.put(None)  # Shuts down worker

    # Process the job
    p._ffmpeg_job(p._input_q, status_q)
    status_q.put.assert_has_calls([
        mock.call((60, "started")),
        mock.call((60, 0)),
        mock.call((60, 50)),
        mock.call((60, 100)),
        mock.call((60, "finished")),
        mock.call((72, "started")),
        mock.call((72, 0)),
        mock.call((72, 50)),
        mock.call((72, 100)),
        mock.call((72, "finished"))
    ])


@mock.patch("m4b_util.helpers.ParallelFFmpeg.FFProgress")
def test_ffmpeg_job_fail(ff, processor):
    """Run a single worker job, with a failing task."""
    # Mock out the run command so we don't have to be dependent on system configurations.
    def ffrun():
        raise RuntimeError("FAIL!")
    ff().run.side_effect = ffrun
    p, _ = processor
    status_q = mock.MagicMock()

    # Add jobs
    p._input_q.put((84, ["should-fail"]))
    p._input_q.put(None)  # Shuts down worker

    # Process the job
    p._ffmpeg_job(p._input_q, status_q)
    status_q.put.assert_has_calls([
        mock.call((84, "started")),
        mock.call((84, "failed")),
    ])


def test_process_empty_tasklist(processor):
    """Return none if task arg is empty."""
    p, _ = processor
    assert p.process(list()) is None


@mock.patch("m4b_util.helpers.ParallelFFmpeg.Progress")
def test_process(progress):
    """Run a multiprocess job."""
    p = ParallelFFmpeg("Testing")

    # Create tasklist.
    # Because _ffmpeg_job will be run in separate processes, we can't mock out FFProgress. This means we do, actually,
    # need a real command.
    tasks = list()
    for i in range(25):
        tasks.append({
            "name": str(i),
            "command": ["ffprobe", "-version"]
        })

    # Process the job.
    # We can't check anything that happens in _ffmpeg_job because it is run in a separate process.
    assert p.process(tasks)
    progress().start.assert_called()
    progress().add_task.assert_has_calls([
        mock.call("[cyan]Testing", total=25)
    ])
    progress().stop.assert_called()


@mock.patch("m4b_util.helpers.ParallelFFmpeg.Progress")
def test_process_crashing_tasks(progress):
    """Run a job with tasks that crash.

    In a previous version, crashing tasks, or too few tasks would lead to an infinite loop.
    """
    p = ParallelFFmpeg("Testing")

    # Create tasklist.
    tasks = list()
    for i in range(2):
        tasks.append({
            "name": str(i),
            "command": ["not-a-command"]
        })

    # Process the job.
    assert p.process(tasks)
    progress().start.assert_called()
    progress().add_task.assert_has_calls([
        mock.call("[cyan]Testing", total=2)
    ])
    progress().stop.assert_called()



"""Chapter Finder."""
from dataclasses import dataclass
from pathlib import Path

from m4b_util.helpers import ffprobe
from .SegmentData import SegmentData


@dataclass
class ChapterFinder:
    """Reads chapter metadata from a file and generates a list of matching SegmentData's."""
    input_path: Path
    start_time: float = 0.0
    end_time: float = 0.0

    def find(self):
        """Read chapter metadata into an object."""
        start_time = self.start_time or 0
        end_time = self.end_time or 100000000000000.0
        probe = ffprobe.run_probe(self.input_path)
        if probe is None:
            return []  # If we can't read the file, then we didn't find any chapters.
        chapter_list = list()
        for chapter in probe.chapters:
            title = chapter.get("tags", dict()).get("title")
            if float(chapter['start_time']) >= start_time and float(chapter['end_time']) <= end_time:
                chapter_list.append(SegmentData(
                    id=chapter['id'],
                    start_time=float(chapter['start_time']),
                    end_time=float(chapter['end_time']),
                    title=title
                ))
        return chapter_list



"""Silence Finder."""
from dataclasses import dataclass, field
from pathlib import Path
import re

from m4b_util.helpers import ffprogress
from .SegmentData import SegmentData


@dataclass
class SilenceFinder:
    """Finds silence in a file and generates a list of SegmentData's representing the non-silence portions."""
    input_path: Path
    start_time: float = 0.0
    end_time: float = 0.0
    silence_duration: float = 3.0
    silence_threshold: int = -35
    _ffoutput: list = field(init=False, repr=False)

    def find(self):
        self._run_silencedetect()
        times = self._get_segment_times()
        if not times:
            return list()

        # Generate SegmentData list
        retval = list()
        for i, (start_time, end_time) in enumerate(times):
            retval.append(SegmentData(
                id=i,
                start_time=start_time,
                end_time=end_time,
            ))
        return retval

    def _run_silencedetect(self):
        """Run the silencedetect filter from ffmpeg and return the output lines."""
        # Run ffmpeg's silencedetect filter
        cmd = ["ffmpeg", "-i", self.input_path]
        if self.start_time:
            cmd.extend(["-ss", str(self.start_time)])
        if self.end_time:
            cmd.extend(["-to", str(self.end_time)])
        cmd.extend([
            "-filter_complex",
            f"[0]silencedetect=d={self.silence_duration}:n={self.silence_threshold}dB[s0]",
            "-map", "[s0]",
            "-f", "null",
            "-"
        ])
        ff = ffprogress.run(cmd, "Detecting Silence")

        # Check to see if ffmpeg exited abnormally
        if ff is None:
            print("[bold red]Error:[/] Could not determine segment times.")
            exit(1)
        self._ffoutput = ff.output.splitlines()

    def _get_segment_times(self):
        """Compute non-silence segment start and end times."""
        silence_start_re = re.compile(r' silence_start: (?P<start>[0-9]+(\.?[0-9]*))$')
        silence_end_re = re.compile(r' silence_end: (?P<end>[0-9]+(\.?[0-9]*)) ')
        total_duration_re = re.compile(
            r'size=[^ ]+ time=(?P<hours>[0-9]{2}):(?P<minutes>[0-9]{2}):(?P<seconds>[0-9\.]{5}) bitrate=')
        # Segments start when silence ends, and segments end when silence starts.
        segment_starts = []
        segment_ends = []
        start_time = self.start_time or 0
        end_time = 10000000.0
        for line in self._ffoutput:
            # Check for silence start
            silence_start_match = silence_start_re.search(line)
            if silence_start_match:
                segment_ends.append(float(silence_start_match.group('start')))
                if len(segment_starts) == 0:
                    # Started with non-silence.
                    segment_starts.append(start_time)
                continue

            # Check for silence end
            silence_end_match = silence_end_re.search(line)
            if silence_end_match:
                segment_starts.append(float(silence_end_match.group('end')))
            total_duration_match = total_duration_re.search(line)

            # Check for duration
            if total_duration_match:
                hours = int(total_duration_match.group('hours'))
                minutes = int(total_duration_match.group('minutes'))
                seconds = float(total_duration_match.group('seconds'))
                end_time = hours * 3600 + minutes * 60 + seconds

        if len(segment_starts) == 0:
            # No silence found.
            return None

        if len(segment_starts) > len(segment_ends):
            # Finished with non-silence.
            segment_ends.append(end_time)

        output = list(zip(segment_starts, segment_ends))

        return output




"""Binder Class."""
from pathlib import Path
import re
import shutil
import subprocess
from tempfile import mkdtemp

from rich import print
from rich.status import Status

from ..helpers import ffprobe, ffprogress
from ..helpers.ParallelFFmpeg import ParallelFFmpeg


class Binder:
    """Read and store all the info we need to bind the audiobook."""

    def __init__(self):
        """Set default values."""
        # Config Related
        self.use_filename = False
        self.decode_durations = False
        self.keep_temp_files = False

        # File Related
        self.files = list()
        self.temp_files = list()
        self.temp_path = Path(mkdtemp(prefix="m4b-util"))
        self.final_file = self.temp_path / "finished.m4b"
        self.output_name = None
        self.output_dir = None

        # Metadata related
        self.chapters = list()
        self.title = None
        self.author = None
        self.date = None
        self.cover = None
        # Set duration to a big number, (hopefully larger than our actual duration)
        # just in case we try to write chapters without knowing the final duration.
        self.duration = 100000000000

    @property
    def output_path(self):
        """Calculate output path based on what we currently know."""
        # Use a default is self.output_name is not set.
        name = self.output_name or f"{self.author} - {self.title}.m4b"

        # Make sure it ends with .m4b
        if not name.endswith(".m4b"):
            name = f"{name}.m4b"

        # Use current dir if self.output_dir is not set.
        if self.output_dir:
            final_path = Path(self.output_dir)
        else:
            final_path = Path()
        return final_path / name

    @property
    def metadata(self):
        """Generate the metadata text when requested, since it depends on the current state of the object."""
        # Start by writing the global header
        metadata = (";FFMETADATA1\n"
                    "major_brand=M4A\n"
                    "minor_version=512\n"
                    "compatible_brands=M4A isomiso2\n"
                    f"title={self.title}\n"
                    f"artist={self.author}\n"
                    f"album={self.title}\n"
                    f"date={self.date}\n"
                    "genre=Audiobook\n"
                    )
        # Add one last entry to our chapter list, to mark the end of the file. This will not get written to the file.
        tmp_chapters = self.chapters.copy()
        tmp_chapters.append({"title": "The Very End of the Book", "startTime": self.duration})

        # Then add chapters to metadata
        for i in range(len(tmp_chapters) - 1):
            title = tmp_chapters[i]['title']
            start = tmp_chapters[i]['startTime']
            end = tmp_chapters[i + 1]['startTime'] - 1
            metadata += ("[CHAPTER]\n"
                         "TIMEBASE=1/1000\n"
                         f"START={start}\n"
                         f"END={end}\n"
                         f"title={title}\n"
                         )
        return metadata

    def bind(self):
        """Bind together an audiobook, based on what we know."""
        if self.keep_temp_files:
            print(f"[yellow]Info:[/] Temp files can be found at [yellow]{self.temp_path}[/]")

        # Make sure we actually have files to convert
        if len(self.files) == 0:
            print("[bold red]Error:[/] No input files found.")
            return False

        # First, convert all audios to m4a, in parallel.
        p = ParallelFFmpeg("Converting files to m4a")
        tasks = list()
        for file in self.files:
            # Determine what to call the output file
            out_m4a = self.temp_path / f"{file.stem}.m4a"

            # Record the file for concatenation later
            self.temp_files.append(out_m4a)

            # Write the FFMPEG command
            cmd = ["ffmpeg", "-i", file, "-filter_complex", "[0:a]asetpts=N/SR/TB[s0]", "-map", "[s0]",
                   "-c:a", "aac", out_m4a, "-y"]
            tasks.append({
                "name": file.stem,
                "command": cmd
            })
        p.process(tasks)

        # Scan the files we just converted for info
        self._scan_files(self.temp_files)

        # Next, concatenate those m4a files into an m4b
        long_file = self._concatenate_all_audio_files()

        # Now, add chapter self
        coverless_file = self._add_chapter_info(long_file)

        # Add cover artwork if specified, otherwise just copy the previous output
        if self.cover:
            covered_file = self._add_cover(coverless_file)
            shutil.move(covered_file, self.final_file)
        else:
            shutil.move(coverless_file, self.final_file)

        # Copy the output into place
        print("[cyan]Copying output.")
        shutil.copy(self.final_file, self.output_path)

        # Clean up
        if not self.keep_temp_files:
            shutil.rmtree(self.temp_path)

        return True

    def _scan_files(self, files):
        """Scan a list of files and update based on them.

        :param files: List of files to scan.
        """
        # Set up variables
        time_counter = 0
        segment_counter = 0

        # Start our status tracker
        print("[cyan]Collecting file data...[/]")
        file_scan_status = Status("Starting")
        file_scan_status.start()

        # Scan all the files
        for file in files:
            file_scan_status.update(f"Scanning {file.name}")
            # Count which segment we are on. Used in case the chapter has no title.
            segment_counter += 1

            # Run ffprobe and parse the output.
            probe = ffprobe.run_probe(file)
            if not probe or probe.audio is None:
                print(f"[bold yellow]Warning:[/] Unable to parse '[bold white]{file}[/]'. Skipping.")
                continue

            # Attempt to fill any missing metadata
            chapter_title = None
            if probe.tags:  # pragma: no branch - None of our test files hit this, but some other audio formats might.
                self.title = self.title or probe.tags.get('album')
                self.author = self.author or probe.tags.get('artist')
                self.date = self.date or probe.tags.get('date')
                chapter_title = probe.tags.get('title')

            if self.use_filename:
                chapter_title = file.stem
            if not chapter_title:
                chapter_title = segment_counter
            self.chapters.append({"title": chapter_title, "startTime": time_counter})

            # Add our duration to the counter for use as the next chapter's start point
            try:
                time_counter += self._get_duration(file)
            except RuntimeError:
                print(f"[yellow]Warning:[/] Failed to determine duration of '[bold white]{file}[/]'. Ignoring.")
                self.chapters.pop()

        # Mark the total duration of all files.
        self.duration = time_counter

        # Shut off the status tracker and alert the user
        file_scan_status.stop()
        print("[green]File scan complete.")

    def _get_duration(self, file):
        """Determine the duration of a file.

        If self.decode_durations is set, run the file through ffmpeg to get the time.
        Otherwise, get it from the metadata.

        :param file: The audio file to run through ffmpeg

        :return The duration of the file, in milliseconds
        """
        duration = None
        if self.decode_durations:
            # Setup variables
            time_regex = re.compile(r"time=(?P<hour>\d{2}):(?P<min>\d{2}):(?P<sec>\d{2})\.(?P<ms>\d{2})")
            time_match = None

            # Run the decoder
            cmd = ["ffmpeg", "-i", file, "-loglevel", "info", "-nostats", "-f", "null", "-"]
            p = subprocess.run(cmd, capture_output=True)

            # Search for time messages
            for line in p.stderr.decode('utf-8', errors="replace").splitlines():
                match = time_regex.search(line)
                if match:
                    time_match = match  # Only keep the most recent

            # Set the duration if we found any
            if time_match:
                duration = time_match.groupdict()
                duration = ffprogress.to_ms(**duration)
            else:
                print(f"[bold yellow]Warning:[/] Unable to find duration of '[bold white]{file.name}[/]' "
                      "during decoding. Falling back to metadata.")

        if not duration:  # Use the metadata as given
            probe = ffprobe.run_probe(file)
            if not probe or probe.audio is None:
                raise RuntimeError("Could not get audio stream.")
            try:
                seconds = float(probe.audio.get('duration'))
            except TypeError:
                raise RuntimeError("Cannot parse duration listed in file.")
            duration = int(seconds * 1000)  # Convert to milliseconds

        return duration

    def _concatenate_all_audio_files(self):
        """Bring all the input files, now m4a, and put them into an m4b."""
        # Start by writing the filelist to a temp file.
        filelist_file = self.temp_path / "filelist"
        with open(filelist_file, 'w') as f:
            for file in self.temp_files:
                # Single quotes in filenames need to be replaced with `'\''` to work with ffmpeg
                # https://superuser.com/questions/787064/filename-quoting-in-ffmpeg-concat
                safer_filename = str(file).replace("'", "'\\''")
                f.write(f"file '{safer_filename}'\n")

        # Then use that file to concatenate everything.
        output_file = self.temp_path / "long.m4a"
        cmd = ["ffmpeg", "-f", "concat", "-safe", "0", "-i", filelist_file, "-c", "copy", output_file]
        self._run_ffmpeg(cmd, "Combining Audio Files.")

        # Cleanup, if applicable
        if not self.keep_temp_files:
            for file in self.temp_files:
                file.unlink()

        return output_file

    def _add_chapter_info(self, input_file):
        """Add chapter info to an m4b file."""
        # Write the metadata to a temp file.
        metadata_file = self.temp_path / "ffmetadata"
        with open(metadata_file, 'w') as f:
            print(self.metadata, file=f)

        output_file = self.temp_path / "coverless.m4b"
        cmd = ["ffmpeg", "-y", "-i", input_file, "-i", metadata_file, "-map_metadata", "1", "-map_chapters", "1",
               "-c", "copy", output_file]
        self._run_ffmpeg(cmd, "Converting to .m4b")

        # Cleanup, if applicable
        if not self.keep_temp_files:
            input_file.unlink()

        return output_file

    def _add_cover(self, input_file):
        """Add a cover to an m4b."""
        output_file = self.temp_path / "covered.m4b"
        cmd = ["ffmpeg", "-y", "-i", input_file, "-i", self.cover, "-map", "0:a", "-map", "1",
               "-c", "copy", "-disposition:v:0", "attached_pic", output_file]
        self._run_ffmpeg(cmd, "Adding Cover")

        # Cleanup, if applicable
        if not self.keep_temp_files:
            input_file.unlink()

        return output_file

    def _run_ffmpeg(self, cmd, message):
        if not ffprogress.run(cmd, message):
            print("[bold red]Error:[/] ffmpeg failed.")
            if not self.keep_temp_files:
                shutil.rmtree(self.temp_path)
            exit(1)



"""Test Split-by-silence."""
from pathlib import Path

import pytest

from m4b_util.split import SegmentData, SilenceFinder


fake_input = Path("Not-a-real-file")


def test_get_segment_times():
    """Extract segment times from ffmpeg output."""
    lines = [
        "[silencedetect @ 0x12860aa70] silence_start: 2.5000",
        "size=N/A time=00:00:25.00 bitrate=N/A speed= 836x",
        "[silencedetect @ 0x12860aa70] silence_end: 5.0000 | silence_duration: 2.5000",
        "[silencedetect @ 0x12860aa70] silence_start: 7.5000",
        "[silencedetect @ 0x12860aa70] silence_end: 10.000 | silence_duration: 2.5000",
        "[silencedetect @ 0x12860aa70] silence_start: 12.5000",
        "[silencedetect @ 0x12860aa70] silence_end: 15.0000 | silence_duration: 2.5000",
        "[silencedetect @ 0x12860aa70] silence_start: 17.5000",
        "[silencedetect @ 0x12860aa70] silence_end: 20.000 | silence_duration: 2.5000",
    ]
    expected = [(0.0, 2.5), (5.0, 7.5), (10.0, 12.5), (15.0, 17.5), (20.0, 25.0)]
    sf = SilenceFinder(fake_input)
    sf._ffoutput = lines
    assert sf._get_segment_times() == expected


def test_get_segment_times_non_zero_start_time():
    """Extract segment times from ffmpeg output, starting at a non-zero time."""
    lines = [
        "[silencedetect @ 0x12860aa70] silence_end: 5.0000 | silence_duration: 2.5000",
        "size=N/A time=00:00:25.00 bitrate=N/A speed= 836x",
        "[silencedetect @ 0x12860aa70] silence_start: 7.5000",
        "[silencedetect @ 0x12860aa70] silence_end: 10.000 | silence_duration: 2.5000",
        "[silencedetect @ 0x12860aa70] silence_start: 12.5000",
        "[silencedetect @ 0x12860aa70] silence_end: 15.0000 | silence_duration: 2.5000",
        "[silencedetect @ 0x12860aa70] silence_start: 17.5000",
        "[silencedetect @ 0x12860aa70] silence_end: 20.000 | silence_duration: 2.5000",
    ]
    expected = [(5.0, 7.5), (10.0, 12.5), (15.0, 17.5), (20.0, 25.0)]
    sf = SilenceFinder(fake_input, start_time=1.25)
    sf._ffoutput = lines
    assert sf._get_segment_times() == expected


def test_get_segment_times_detect_end_first():
    """Handle segment times even if silence_end comes first."""
    lines = [
        "[silencedetect @ 0x12860aa70] silence_end: 5.0000 | silence_duration: 2.5000",
        "size=N/A time=00:00:25.00 bitrate=N/A speed= 836x",
        "[silencedetect @ 0x12860aa70] silence_start: 7.5000",
        "[silencedetect @ 0x12860aa70] silence_end: 10.000 | silence_duration: 2.5000",
        "[silencedetect @ 0x12860aa70] silence_start: 12.5000",
        "[silencedetect @ 0x12860aa70] silence_end: 15.0000 | silence_duration: 2.5000",
        "[silencedetect @ 0x12860aa70] silence_start: 17.5000",
        "[silencedetect @ 0x12860aa70] silence_end: 20.000 | silence_duration: 2.5000",
        "[silencedetect @ 0x12860aa70] silence_start: 22.5000",
    ]
    expected = [(5.0, 7.5), (10.0, 12.5), (15.0, 17.5), (20.0, 22.5)]
    sf = SilenceFinder(fake_input, start_time=1.25)
    sf._ffoutput = lines
    assert sf._get_segment_times() == expected


def test_get_segment_times_empty():
    """Notice when we don't have any silence to extract."""
    expected = None
    sf = SilenceFinder(fake_input, start_time=1.25)
    sf._ffoutput = [""]
    assert sf._get_segment_times() == expected


def test_run_silencedetect(silences_file_path):
    """Get ffmpeg output from silencedetect filter."""
    expected = [
        "silence_start: 2.49979",
        "silence_end: 5.0001 | silence_duration: 2.50031",
        "silence_start: 7.49992",
        "silence_end: 10.0001 | silence_duration: 2.50017",
        "silence_start: 12.4999",
        "silence_end: 15.0001 | silence_duration: 2.50012",
        "silence_start: 17.5",
        "size=N/A time=00:00:20.01 bitrate=N/A",
    ]
    sf = SilenceFinder(
        input_path=silences_file_path,
        silence_duration=0.25
    )
    sf._run_silencedetect()
    for item in expected:
        match_list = [line for line in sf._ffoutput if item in line]
        assert len(match_list) == 1


def test_run_silencedetect_fail(tmp_path, capsys):
    """Get ffmpeg output from a failed silencedetect filter."""
    fake_file = tmp_path / "Not-a-real-file.mp3"
    open(fake_file, 'a').close()
    sf = SilenceFinder(fake_file)
    with pytest.raises(SystemExit) as e:
        sf._run_silencedetect()
    assert (e.value.code == 1)
    output = capsys.readouterr()
    assert "Could not determine segment times." in output.out


def test_run_silencedetect_no_silence(silences_file_path):
    """Get ffmpeg output from a no-silence-detected run."""
    unexpected = [
        "silence_start",
        "silence_end",
        "silence_duration",
    ]
    sf = SilenceFinder(silences_file_path)
    sf._run_silencedetect()
    for item in unexpected:
        match_list = [line for line in sf._ffoutput if item in line]
        assert len(match_list) == 0


def test_run_silencedetect_variable_start_end_times(silences_file_path):
    """Get ffmpeg output from silencedetect filter."""
    expected = [
        "silence_end: 5.0001 | silence_duration: 2.50031",
        "silence_start: 7.49992",
        "silence_end: 10.0001 | silence_duration: 2.50017",
        "silence_start: 12.4999",
        "silence_end: 15.0001 | silence_duration: 2.50012",
        "size=N/A time=00:00:13.00 bitrate=N/A",
    ]
    sf = SilenceFinder(
        input_path=silences_file_path,
        silence_duration=0.25,
        start_time=3,
        end_time=16
    )
    sf._run_silencedetect()
    for item in expected:
        match_list = [line for line in sf._ffoutput if item in line]
        assert len(match_list) == 1


def test_find(silences_file_path):
    """Find silences in a file."""
    expected = [
        SegmentData(id=0, start_time=00.0000, end_time=2.49979),
        SegmentData(id=1, start_time=05.0001, end_time=7.49992),
        SegmentData(id=2, start_time=10.0001, end_time=12.4999),
        SegmentData(id=3, start_time=15.0001, end_time=17.5000),
        SegmentData(id=4, start_time=20.0107, end_time=20.0100)
    ]
    actual = SilenceFinder(silences_file_path, silence_duration=0.25).find()
    assert actual == expected



from unittest import mock

from m4b_util.split import SegmentData, Splitter
from m4b_util.helpers import ffprobe
import testhelpers


def test_splitter(silences_file_path, tmp_path):
    """Split a file into four parts"""
    output_path = tmp_path / "output"
    segment_list = [
        SegmentData(id=0, start_time=0.0, end_time=2.5),
        SegmentData(id=1, start_time=2.5, end_time=5.0),
        SegmentData(id=2, start_time=5.0, end_time=7.5),
        SegmentData(id=3, start_time=7.5, end_time=10.),

    ]
    expected_files = [
        "segment_0000.mp3",
        "segment_0001.mp3",
        "segment_0002.mp3",
        "segment_0003.mp3",
    ]
    Splitter(
        input_path=silences_file_path,
        output_dir_path=output_path,
        segment_list=segment_list,
    ).split()
    testhelpers.check_output_folder(output_path=output_path, expected_files=expected_files)


def test_alternate_output_pattern(silences_file_path, tmp_path):
    """Split a file into four parts, with custom naming rules"""
    output_path = tmp_path / "output"
    segment_list = [
        SegmentData(id=0, start_time=0.0, end_time=2.5),
        SegmentData(id=1, start_time=2.5, end_time=5.0, title="Frist"),
        SegmentData(id=2, start_time=5.0, end_time=7.5),
        SegmentData(id=3, start_time=7.5, end_time=10., title="Secnod"),

    ]
    expected_files = [
        "00 - None.mp3",
        "01 - Frist.mp3",
        "02 - None.mp3",
        "03 - Secnod.mp3",
    ]
    Splitter(
        input_path=silences_file_path,
        output_dir_path=output_path,
        segment_list=segment_list,
        output_pattern="{i:02d} - {title}.mp3"
    ).split()
    testhelpers.check_output_folder(output_path=output_path, expected_files=expected_files)


def test_overlapping_output_names(silences_file_path, tmp_path):
    """Overwrite files, if the naming pattern causes collisions."""
    def check_func(input_file_path):
        """Check for correct file length."""
        probe = ffprobe.run_probe(input_file_path)
        assert probe.audio['duration'] == "5.040000"
    output_path = tmp_path / "output"
    segment_list = [
        SegmentData(id=0, start_time=0.0, end_time=2.5),
        SegmentData(id=1, start_time=2.5, end_time=5.0, title="Frist"),
        SegmentData(id=2, start_time=5.0, end_time=5.5),
        SegmentData(id=3, start_time=5.5, end_time=10.5, title="Secnod"),

    ]
    expected_files = [
        "Collided_File.mp3",
    ]
    Splitter(
        input_path=silences_file_path,
        output_dir_path=output_path,
        segment_list=segment_list,
        output_pattern="Collided_File.mp3",
    ).split()

    testhelpers.check_output_folder(output_path=output_path, expected_files=expected_files, check_func=check_func)


def test_title_metadata(silences_file_path, tmp_path):
    """Split a file into four parts"""
    def check_func(input_file_path):
        probe = ffprobe.run_probe(input_file_path)
        assert input_file_path.name == probe.tags["title"]
    output_path = tmp_path / "output"
    segment_list = [
        SegmentData(id=0, start_time=0.0, end_time=2.5, title="segment_0000"),
        SegmentData(id=1, start_time=2.5, end_time=5.0, title="segment_0001"),
        SegmentData(id=2, start_time=5.0, end_time=7.5, title="segment_0002"),
        SegmentData(id=3, start_time=7.5, end_time=10., title="segment_0003"),

    ]
    expected_files = [
        "segment_0000.mp3",
        "segment_0001.mp3",
        "segment_0002.mp3",
        "segment_0003.mp3",
    ]
    Splitter(
        input_path=silences_file_path,
        output_dir_path=output_path,
        segment_list=segment_list,
    ).split()
    testhelpers.check_output_folder(output_path=output_path, expected_files=expected_files)




import argparse
from pathlib import Path
import sys

from rich import print

from .SilenceFinder import SilenceFinder
from .ChapterFinder import ChapterFinder
from .Splitter import Splitter


def _parse_split_silence_args():
    parser = argparse.ArgumentParser(
        prog="m4b-util split",
        description='Split media into segments.'
    )
    parser.add_argument('mode', help='modes: s - silence, c - chapters')
    parser.add_argument('input_file', help='Input filename')
    parser.add_argument('-e', "--end-time", type=float, help='End time (seconds)')
    parser.add_argument('-m', "--minimum_segment_time", type=float, default=1.0,
                        help='Smallest segment size to consider, in seconds.')
    parser.add_argument('-o', "--output-dir", type=str, help="Directory to place output.")
    parser.add_argument('-p', '--output-pattern', type=str, default="segment_{i:04d}.mp3",
                        help="Output filename pattern (e.g. `segment_{i:04d}.mp3`), use '{i}' for sequence and "
                             "'{title}' for chapter title.")
    parser.add_argument('-s', "--start-time", type=float, help='Start time (seconds)')
    parser.add_argument("--silence-threshold", default=-35, type=int, help='Silence threshold (in dB)')
    parser.add_argument("--silence-duration", default=3.0, type=float, help='Silence duration')

    return parser.parse_args(sys.argv[2:])


def run():
    """Split an audio file into pieces, based on silence."""
    args = _parse_split_silence_args()
    input_path = Path(args.input_file)
    segment_list = None

    # Step 1: Look for where the splits should be.
    if args.mode.lower() in ["s", "silence", "silences"]:
        segment_list = SilenceFinder(
            input_path=input_path,
            start_time=args.start_time,
            end_time=args.end_time,
            silence_duration=args.silence_duration,
            silence_threshold=args.silence_threshold,
        ).find()
    elif args.mode.lower() in ["c", "chapter", "chapters"]:
        segment_list = ChapterFinder(
            input_path=input_path,
            start_time=args.start_time,
            end_time=args.end_time,
        ).find()
    else:
        print(f"[bold red]Error:[/] Unexpected mode '{args.mode}'")
        exit(1)

    if not segment_list:
        print("[bold red]Error:[/] No segments found.")
        exit(1)

    # Step 2: Filter out too-small segments
    # Use .copy() so that we can modify the original without throwing off the iterator
    for segment in segment_list.copy():
        if segment.end_time - segment.start_time < args.minimum_segment_time:
            segment_list.remove(segment)
    if len(segment_list) == 0:
        print("[bold red]Error:[/] Not enough segments found.")
        exit(1)

    # Step 3: Split the original file
    print(f"[yellow]Info:[/] Found [yellow]{len(segment_list)}[/] segments")
    if args.output_dir:
        output_path = Path(args.output_dir)
    else:
        output_path = Path()
    Splitter(
        input_path=input_path,
        output_dir_path=output_path,
        output_pattern=args.output_pattern,
        segment_list=segment_list
    ).split()





"""ffprogress tests."""
import io
from unittest import mock

import pytest

from m4b_util.helpers import ffprobe, ffprogress


def test_to_ms():
    """Convert hr:min:sec.ms to milliseconds."""
    assert ffprogress.to_ms() == 0
    assert ffprogress.to_ms(sec=5, ms=555) == 5555
    assert ffprogress.to_ms(hour=5, min=5, sec=5, ms=555) == 18305555


@mock.patch("subprocess.Popen")
def test_ffprogress_class(popen):
    """Use the FFProgress class."""
    # Make sure output is initialized to None
    ff = ffprogress.FFProgress(["mocked_cmd"])
    assert ff.output is None

    # Set up Mock
    type(popen()).stdout = mock.PropertyMock(side_effect=[
        None,
        io.BytesIO(b"Duration: 01:00:00.00"),
        None,
        io.BytesIO(b""),
        io.BytesIO(b"out_time=00:00:00.00"),
        io.BytesIO(b"out_time=00:15:00.00"),
        io.BytesIO(b"out_time=00:30:00.00"),
        io.BytesIO(b"out_time=00:45:00.00"),
        io.BytesIO(b"out_time=01:00:00.00"),
        io.BytesIO(b"")
    ])
    popen().returncode = 0
    popen().poll.side_effect = [None, 0]

    expected = [0, 0, 25, 50, 75, 100, 100]
    for percent in ff.run():
        assert percent == expected.pop(0)


@mock.patch("subprocess.Popen")
def test_ffprogress_class_fail(popen):
    """Handle a command that fails."""
    ff = ffprogress.FFProgress(["mocked_cmd"])

    # Set up Mock
    popen().stdout = io.BytesIO(b"")
    popen().returncode = -1
    popen().poll.return_value = 0

    with pytest.raises(RuntimeError) as e:
        for percent in ff.run():
            assert percent == 0
    assert ("Error running command ['mocked_cmd']:" in str(e.value))


def test_run(tmp_path, capsys):
    """Use the helper run function."""
    output_path = tmp_path / "file_with_silences.m4a"
    cmd = ["ffmpeg",
           "-f", "lavfi", "-i", "sine=frequency=110:sample_rate=48000:duration=2.5",
           "-f", "lavfi", "-i", "sine=frequency=220:sample_rate=48000:duration=2.5",
           "-f", "lavfi", "-i", "sine=frequency=330:sample_rate=48000:duration=2.5",
           "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5",
           "-filter_complex",
           "[0]apad=pad_dur=2.5[s0],[1]apad=pad_dur=2.5[s1],[2]apad=pad_dur=2.5[s2],[3]apad=pad_dur=2.5[s3],"
           "[s0][s1][s2][s3]concat=n=4:v=0:a=1[s4]", "-map", "[s4]",
           output_path
           ]
    ffprogress.run(cmd, "Testing progress message")
    output = capsys.readouterr()
    assert ("Testing progress message" in output.out)

    # Check that the output matches what we expect
    probe = ffprobe.run_probe(output_path)
    assert probe.audio
    assert probe.format['duration'] == "20.000000"


def test_run_fail(capsys):
    """Handle a crashing command when using the helper run function."""
    cmd = ["ffmpeg", "-not-a-real-option"]
    ffprogress.run(cmd, "Testing progress message")
    output = capsys.readouterr()
    assert ("Something went wrong with ffmpeg" in output.out)


def test_run_no_printing_errors():
    """Handle a crashing command with print_errors turned off."""
    cmd = ["ffmpeg", "-not-a-real-option"]
    with pytest.raises(RuntimeError) as e:
        ffprogress.run(cmd, print_errors=False)
    assert "Error running command" in str(e.value)



"""Test the Bind subcommand."""
from unittest.mock import patch

import pytest

from m4b_util.__main__ import allowed_commands, main as m4b_main


def _run_main_cmd(arg_list, expected_exit=0):
    argv_patch = ["m4b-util"]
    argv_patch.extend(arg_list)

    with patch("sys.argv", argv_patch):
        with pytest.raises(SystemExit) as e:
            m4b_main()
        assert e.value.code == expected_exit


def test_print_version(capsys):
    """Show the order the files would be bound in, if asked."""
    _run_main_cmd(["version"])

    output = capsys.readouterr()
    expected_output = "m4b-util, Version"  # Don't specify the exact version, so we don't have to update it every tag.
    assert expected_output in output.out


def test_unrecognized_command(capsys):
    """Throw an error on an unrecognized sub-command."""
    _run_main_cmd(["definitely-not-a-real-command"], -1)

    output = capsys.readouterr()
    expected_output = "Unrecognized command"
    assert expected_output in output.out


def test_all_commands_help(capsys):
    """Display help menu from all commands."""
    for command in allowed_commands:
        _run_main_cmd([command, "--help"])



"""Helper functions for running ffprobe."""
import json
from subprocess import run


class Probe:
    """Holds the results from an ffprobe run."""
    def __init__(self, ffprobe_output):
        """Set up attributes."""
        self.data = json.loads(ffprobe_output)
        self.streams = self.data.get('streams', None)
        self.audio = next(
            (stream for stream in self.data.get('streams', list()) if stream.get('codec_type') == 'audio'), None
        )
        self.format = self.data.get('format')
        self.tags = self.data.get('format', dict()).get('tags', None)
        self.chapters = self.data.get('chapters')


def run_probe(file):
    """Run ffprobe on the specified file and return a JSON representation of the output."""
    cmd = ["ffprobe", "-show_format", "-show_streams", "-show_chapters", "-of", "json", "-i", file]

    p = run(cmd, capture_output=True)
    if p.returncode != 0:
        return None
    return Probe(p.stdout.decode('utf-8'))



"""Helper functions for running FFMPEG."""
import re
import subprocess

from rich.progress import Progress, SpinnerColumn


def to_ms(hour=0, min=0, sec=0, ms=0):
    """Convert from hrs:min:sec.ms to straight milliseconds."""
    result = (int(hour) * 60 * 60 * 1000) + (int(min) * 60 * 1000) + (int(sec) * 1000) + int(ms)
    return result


class FFProgress:
    """Represents a single run of an ffmpeg command."""
    DUR_REGEX = re.compile(r"Duration: (?P<hour>\d{2}):(?P<min>\d{2}):(?P<sec>\d{2})\.(?P<ms>\d{2})")
    TIME_REGEX = re.compile(r"out_time=(?P<hour>\d{2}):(?P<min>\d{2}):(?P<sec>\d{2})\.(?P<ms>\d{2})")

    def __init__(self, cmd) -> None:
        """Initialize the FfmpegProgress class.

        :param cmd: A list of command line elements, e.g. ["ffmpeg", "-i", ...]
        """
        self.cmd = cmd
        self.output = None

    def run(self):
        """Run an ffmpeg command, trying to capture the process output and calculate the duration / progress.

        Yields the progress in percent.
        """
        total_dur = None

        # Make sure the command has -progress and -nostats
        cmd_with_progress = ([self.cmd[0]] + ["-progress", "-", "-nostats"] + self.cmd[1:])

        stdout = []

        p = subprocess.Popen(
            cmd_with_progress,
            stdin=subprocess.PIPE,  # Apply stdin isolation by creating separate pipe.
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=False,
        )

        yield 0

        while True:
            # If there isn't anything new to process, just keep spinning.
            line = p.stdout
            if line is None:
                continue

            # Clean up the line we got from subprocess
            stdout_line = line.readline().decode("utf-8", errors="replace").strip()

            # Check if it is time to exit
            if stdout_line == "" and p.poll() is not None:
                break

            # Add what we got to our internal track of stderr
            stdout.append(stdout_line.strip())
            self.output = "\n".join(stdout)

            # Check out regexes
            if total_dur is None:  # We haven't found the initial duration yet
                total_dur_match = FFProgress.DUR_REGEX.search(stdout_line)
                if total_dur_match:
                    total_dur = total_dur_match.groupdict()
                    total_dur = to_ms(**total_dur)
            else:  # We already know our duration, lets see if we have an update
                progress_time = FFProgress.TIME_REGEX.search(stdout_line)
                if progress_time:
                    elapsed_time = to_ms(**progress_time.groupdict())
                    yield (elapsed_time / total_dur) * 100

        # Throw an exception if ffmpeg didn't exit cleanly
        if p.returncode != 0:
            self.output = "\n".join(stdout)
            raise RuntimeError(f"Error running command {str(self.cmd)}: {self.output}")

        # We've got nothing else to give, so mark it as 100% done
        yield 100


def run(cmd, task_name="Thinking", print_errors=True):
    """Run ffmpeg command and show progress.

    :param cmd: Command to run
    :param task_name: Text to print in front of progress bar
    :param print_errors: Toggle whether to print error messages, or re-throw the exception.

    :return FFProgress instance
    """
    progress = Progress(SpinnerColumn(), *Progress.get_default_columns())
    progress.start()
    ff = FFProgress(cmd)

    try:
        task = progress.add_task(f"[cyan]{task_name}", total=100)
        for percent in ff.run():
            progress.update(task, completed=percent)
    except RuntimeError as e:
        if print_errors:
            progress.console.print("[bold red]Error:[/] Something went wrong with ffmpeg:")
            progress.console.print(e)
            progress.stop()
            return None
        else:
            progress.stop()
            raise e

    progress.update(task, completed=100)
    progress.stop()

    return ff



from subprocess import run

import pytest


@pytest.fixture()
def silences_file_path(tmp_path):
    """File containing multiple segments of silence."""
    output_path = tmp_path / "file_with_silences.m4a"
    cmd = ["ffmpeg",
           "-f", "lavfi", "-i", "sine=frequency=110:sample_rate=48000:duration=2.5",
           "-f", "lavfi", "-i", "sine=frequency=220:sample_rate=48000:duration=2.5",
           "-f", "lavfi", "-i", "sine=frequency=330:sample_rate=48000:duration=2.5",
           "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5",
           "-filter_complex",
           "[0]apad=pad_dur=2.5[s0],[1]apad=pad_dur=2.5[s1],[2]apad=pad_dur=2.5[s2],[3]apad=pad_dur=2.5[s3],"
           "[s0][s1][s2][s3]concat=n=4:v=0:a=1[s4]", "-map", "[s4]",
           output_path
           ]

    run(cmd, capture_output=True, check=True)
    return output_path


@pytest.fixture()
def variable_volume_segments_file_path(tmp_path):
    """File containing multiple segments of louder and softer sounds."""
    output_path = tmp_path / "variable_volume_segments.m4b"
    cmd = ["ffmpeg",
           "-f", "lavfi", "-i", "sine=frequency=110:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=110:sample_rate=48000:duration=2.5,volume=.5",
           "-f", "lavfi", "-i", "sine=frequency=220:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=220:sample_rate=48000:duration=2.5,volume=.5",
           "-f", "lavfi", "-i", "sine=frequency=330:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=330:sample_rate=48000:duration=2.5,volume=.5",
           "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5,volume=.5",
           "-filter_complex",
           "[0][1][2][3][4][5][6][7]concat=n=8:v=0:a=1[s0]", "-map", "[s0]",
           output_path
           ]
    run(cmd, capture_output=True, check=True)
    return output_path


@pytest.fixture()
def chaptered_audio_file_path(tmp_path):
    """File containing multiple chapters."""
    metadata = (
        ";FFMETADATA1\nmajor_brand=M4A\nminor_version=512\ncompatible_brands=M4A isomiso2\n"
        "title=Chaptered Audio\nartist=m4b-util\nalbum=Chaptered Audio\ndate=2022\ngenre=Audiobook\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=0\nEND=2499\ntitle=110Hz - Loud\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=2500\nEND=4999\ntitle=110Hz - Soft\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=5000\nEND=7499\ntitle=220Hz - Loud\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=7500\nEND=9999\ntitle=220Hz - Soft\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=10000\nEND=12499\ntitle=330Hz - Loud\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=12500\nEND=14999\ntitle=330Hz - Soft\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=15000\nEND=17499\ntitle=440Hz - Loud\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=17500\nEND=19999\ntitle=440Hz - Soft\n"
        )
    metadata_file = tmp_path / "ffmetadata"
    with open(metadata_file, 'w') as f:
        print(metadata, file=f)
    output_path = tmp_path / "chaptered_audio.m4b"
    cmd = ["ffmpeg",
           "-i", metadata_file,
           "-f", "lavfi", "-i", "sine=frequency=110:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=110:sample_rate=48000:duration=2.5,volume=.5",
           "-f", "lavfi", "-i", "sine=frequency=220:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=220:sample_rate=48000:duration=2.5,volume=.5",
           "-f", "lavfi", "-i", "sine=frequency=330:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=330:sample_rate=48000:duration=2.5,volume=.5",
           "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5,volume=.5",
           "-filter_complex",
           "[1][2][3][4][5][6][7][8]concat=n=8:v=0:a=1[s0]",
           "-map", "[s0]", "-map_metadata", "0", "-map_chapters", "0",
           output_path
           ]
    run(cmd, capture_output=True, check=True)
    return output_path



"""Segment Data."""
from dataclasses import dataclass


@dataclass
class SegmentData:
    """Represents a single segment from an audio file, such as a chapter or a non-silence area."""
    start_time: float
    end_time: float
    id: int = None
    title: str = None



"""Handle processsing audio files in parallel."""
import multiprocessing as mp
import os
import queue

from rich.progress import Progress, SpinnerColumn

from .ffprogress import FFProgress


class ParallelFFmpeg:
    """Process audio files in parallel."""

    def __init__(self, purpose):
        """Initialize everything.

        :param purpose: Description of overall purpose. Printed in front of the overall progress bar.
        """
        # Store our parameters
        self._purpose = purpose

        # Set up our member variables
        num_cores = os.cpu_count()
        self._input_q = mp.Queue(maxsize=num_cores)
        self._status_q = mp.Queue(maxsize=100)
        self._tasklist = list()
        self.temp_files = list()

        # Progress Tracker
        self.progress = Progress(SpinnerColumn(), *Progress.get_default_columns())

    def _read_status_queue(self):
        """Read info from the status queue and update the display."""
        try:
            master_task = self._tasklist[0]["task_id"]
            (job_id, job_status) = self._status_q.get_nowait()
            task_name = self._tasklist[job_id]["name"]
            if job_status == "started":
                self._tasklist[job_id]["task_id"] = self.progress.add_task(
                    f"[dark_cyan]|- Processing '[white]{task_name}[/]'.",
                    total=100
                )
            elif job_status == "finished":
                self.progress.update(self._tasklist[job_id]["task_id"], completed=100, visible=False)
                self.progress.update(master_task, advance=1)
            elif job_status == "failed":
                name = self._tasklist[job_id]["name"]
                self.progress.console.print(f"[red]Error:[/] Failed to process {name}")
                self.progress.update(self._tasklist[job_id]["task_id"], visible=False)
                self.progress.update(master_task, advance=1)
            else:  # No matching text means it is an update on percentage.
                self.progress.update(self._tasklist[job_id]["task_id"], completed=job_status)
        except queue.Empty:
            pass  # We don't actually care if the finished queue is empty

    def _insert_into_input_q(self, args):
        while True:
            # Check the status
            self._read_status_queue()
            # See if there is space in the queue
            try:
                self._input_q.put_nowait(args)
                break  # No need to keep looping once we have successfully written the queue
            except queue.Full:
                pass  # We will just try again.

    @staticmethod
    def _ffmpeg_job(q, status_q):
        """Multiprocessing worker that runs an ffmpeg command."""
        while True:
            args = q.get()
            if args is None:
                break

            (job_id, cmd) = args

            # Let the outside world know we started
            status_q.put((job_id, "started"))

            try:
                ff = FFProgress(cmd)
                for percent in ff.run():
                    status_q.put((job_id, percent))
                status_q.put((job_id, "finished"))
            # Since we need these workers to stay up until we close them, we will catch all but the most
            # dire exceptions.
            except Exception as e:  # noqa: See comment above.
                status_q.put((job_id, "failed"))
                continue

    def process(self, tasks):
        """Use multiprocessing to run all ffmpeg commands.

        :param tasks: A list of dictionaries with two keys: 'name' and 'command'. Name will be printed by the progress
                      tracker while command is being run.
        """
        # Make sure we have tasks
        if len(tasks) == 0:
            return None

        # Start up our progress tracker
        self.progress.start()

        # Set up our processing pool
        num_cores = os.cpu_count()
        pool = mp.Pool(num_cores, initializer=self._ffmpeg_job, initargs=(self._input_q, self._status_q))

        master_task = self.progress.add_task(
            f"[cyan]{self._purpose}",
            total=len(tasks)
        )
        self._tasklist.append({
            "name": "Master",
            "task_id": master_task
        })

        for task in tasks:
            # Mark a new task as beginning.
            self._tasklist.append({
                "name": task.get("name", "unknown"),
                "task_id": None
            })
            job_id = len(self._tasklist) - 1

            # Send the info to our workers
            args = (job_id, task["command"])
            self._insert_into_input_q(args)

        # Tell the workers we are all done
        for _ in range(num_cores):
            self._insert_into_input_q(None)
        pool.close()

        # Wait for all the jobs to finish
        while not (self._input_q.empty() and self._status_q.empty()):
            self._read_status_queue()
        pool.join()

        # Make sure the master task shows as completed
        # Yes, we need to doublecheck the status_q again, because the process of verifying the status_q is empty
        # and joining the pool is not atomic.
        while not self._status_q.empty():
            self._read_status_queue()  # nocover: This is just a timing thing, and can't be tested reliably.
        self.progress.update(master_task, completed=len(tasks))

        # Shut down the progress tracker
        self.progress.stop()

        return True



"""The Splitter Class."""
from dataclasses import dataclass
from pathlib import Path

from m4b_util.helpers.ParallelFFmpeg import ParallelFFmpeg
from m4b_util.helpers import cover_utils


@dataclass
class Splitter:
    """Holds all the info and methods for splitting a file."""
    input_path: Path
    output_dir_path: Path
    segment_list: list

    output_pattern: str = "segment_{i:04d}.mp3"

    def split(self):
        """Do the actual splitting."""
        cover_utils.extract_cover(self.input_path, self.output_dir_path / "cover.png")

        # Generate task list
        tasks = list()
        for i, segment in enumerate(self.segment_list):
            time = segment.end_time - segment.start_time

            self.output_dir_path.mkdir(exist_ok=True)
            output_path = self.output_dir_path / self.output_pattern.format(i, i=i, title=segment.title)

            cmd = ["ffmpeg", "-ss", str(segment.start_time), "-t", str(time), "-i", self.input_path,
                   "-map", "0:a", "-map_chapters", "-1", "-y"]
            if segment.title:
                cmd.extend(["-metadata", f"title={segment.title}"])
            cmd.append(output_path)

            name = f"Splitting segment {i}"
            if segment.title:
                name += f" - {segment.title}"
            tasks.append({
                "name": name,
                "command": cmd
            })

        # Process splits in parallel
        p = ParallelFFmpeg(f"Splitting '{self.input_path.name}'")
        p.process(tasks)







"""Utility functions relating to cover images."""
from rich import print

from m4b_util.helpers import ffprogress


def extract_cover(input_path, output_path):
    """Dump an audio file's cover image to disk.

    :param input_path: The file to extract the cover from
    :param output_path: Where to put the cover, including name. Must end in .png, .jpg, or ,jpeg
    """
    allowed_extensions = [".png", ".jpg", ".jpeg"]
    if output_path.suffix.lower() not in allowed_extensions:
        raise ValueError(f"Output extension must be one of {allowed_extensions}")
    cmd = ["ffmpeg", "-y", "-i", input_path, output_path]
    try:
        ffprogress.run(cmd, task_name="Extracting Cover.", print_errors=False)
    except RuntimeError:
        print("[yellow]Warning:[/] Unable to extract cover.")



"""Main Script for m4b-util."""
import argparse
import sys

from rich import print

from . import bind, split
from .__version__ import version


def _print_version():
    """Print the current version, then exit."""
    print(f"[green]m4b-util[/], Version '{version}'")
    return 0


# Set up the dictionary of commands. The values are tuples, first the function to run, second the description.
allowed_commands = {
    "bind": (bind.run, "Convert a folder of audio files into an m4b."),
    "split": (split.run, "Split a file into smaller pieces."),
    "version": (_print_version, "Print the program's version.")
}


def main():
    """Run the application."""
    # Write our usage message
    usage = ("m4b-util <command> [<args>]\n\n"
             "Allowed Commands:\n"
             )
    for name, (_, description) in allowed_commands.items():
        usage += f"{name:18}: {description}\n"
    usage += ("\nFor more help with a command, use m4b-util <command> --help\n"
              " \n"
              )

    # Set up argparse
    parser = argparse.ArgumentParser(
        prog="m4b-util",
        usage=usage
    )
    parser.add_argument('command', help='Subcommand to run')
    # parse_args defaults to [1:] for args, but we need to exclude the rest of the args so they can be picked up by
    args = parser.parse_args(sys.argv[1:2])

    # Make sure args.command is always lowercase
    args.command = args.command.lower()

    if args.command not in allowed_commands:
        print("[bold red]Error:[/] Unrecognized command")
        print(parser.print_help())
        exit(-1)

    # Invoke the subcommand
    retcode = allowed_commands[args.command][0]()

    print("[green]Done![/]")
    exit(retcode)


# We don't test coverage for this, since we don't test it directly.
# We just make it simple enough that we can trust it works.
if __name__ == "__main__":  # pragma: no cover
    main()  # pragma: no cover



"""ffprobe tests."""
from m4b_util.helpers import ffprobe


def test_video_only(video_only_file):
    """Read in a video-only file."""
    probe = ffprobe.run_probe(video_only_file)
    assert probe.audio is None
    expected_tags = {
        'compatible_brands': 'isomiso2avc1mp41',
        'major_brand': 'isom',
        'minor_version': '512'
    }
    del probe.tags['encoder']  # Encoder will be different depending on the version of ffmpeg
    assert probe.tags == expected_tags


def test_probe_data(mp3_path):
    """Read an mp3 file."""
    file = mp3_path / "6 - 660Hz.mp3"
    probe = ffprobe.run_probe(file)
    assert probe

    expected = {
        'chapters': [],
        'format': {
            'bit_rate': '64376',
            'duration': '5.040000',
            'format_long_name': 'MP2/3 (MPEG audio layer 2/3)',
            'format_name': 'mp3',
            'nb_programs': 0,
            'nb_streams': 1,
            'probe_score': 51,
            'size': '40557',
            'start_time': '0.023021',
            'tags': {}
        },
        'streams': [{
            'avg_frame_rate': '0/0',
            'bit_rate': '64000',
            'bits_per_sample': 0,
            'channel_layout': 'mono',
            'channels': 1,
            'codec_long_name': 'MP3 (MPEG audio layer 3)',
            'codec_name': 'mp3',
            'codec_tag': '0x0000',
            'codec_tag_string': '[0][0][0][0]',
            'codec_type': 'audio',
            'disposition': {
                'attached_pic': 0,
                'captions': 0,
                'clean_effects': 0,
                'comment': 0,
                'default': 0,
                'dependent': 0,
                'descriptions': 0,
                'dub': 0,
                'forced': 0,
                'hearing_impaired': 0,
                'karaoke': 0,
                'lyrics': 0,
                'metadata': 0,
                'original': 0,
                'still_image': 0,
                'timed_thumbnails': 0,
                'visual_impaired': 0
            },
            'duration': '5.040000',
            'duration_ts': 71124480,
            'index': 0,
            'r_frame_rate': '0/0',
            'sample_fmt': 'fltp',
            'sample_rate': '48000',
            'start_pts': 324870,
            'start_time': '0.023021',
            'time_base': '1/14112000'
        }]
    }
    # Remove items that can change between tests
    del probe.data['format']['filename']
    del probe.data['format']['tags']['encoder']
    assert (probe.data == expected)


def test_probe_blank_file(tmp_path):
    """Return None when run on an un-parsable file."""
    file = tmp_path / "not-really-audio.mp3"
    open(file, 'a').close()
    probe = ffprobe.run_probe(file)
    assert not probe
